//**********************************************************************************************************\
//
//	archive_types.h:	Archive
//
//**********************************************************************************************************/

#define ARCHIVE_CREATE		0	// Sequential write-only.  No read.  No modification of commits.  No skipping around.
#define ARCHIVE_READ		1	// Read only.
#define ARCHIVE_EDIT		2	// Comprehensive edit
#define ARCHIVE_FASTEDIT	3	// Fast and wasteful edit.  Does not recover space for deleted files.

#define ARCHIVE_CREATE_WAD3 9

// archive_pak.c
#define PakFile_Add_File_Pathless(myPack, myPathName) PakFile_Add_File_As (myPack, myPathName, String_SkipPath(myPathName) );


struct pak_access_s*	PakFile_Auto_Open	(const char* myPakFileName); // Wraps PakFile_Open

struct pak_access_s*	PakFile_Open		(const char* myPakFileName, int myFileMode);
struct pak_access_s*	PakFile_Close		(struct pak_access_s* myPack);

fbool					PakFile_Delete_File (struct pak_access_s* myPack, const char* storeAsName);
fbool					PakFile_Pack_File	(struct pak_access_s* myPack, const char* diskFileName, const char* storeAsName);
fbool					PakFile_Rename_File (struct pak_access_s* myPack, const char* storeAsName, const char* newName);
fbool					PakFile_Add_File_As (struct pak_access_s* myPack, const char* diskFileName, const char* storeAsName);
fbool					PakFile_Pack_File_As(struct pak_access_s* myPack, const char* diskFileName, const char* storeAsName); // Replaces or Adds
fbool					PakFile_Pak_Info	(const char* myPakFileName);

int						PakFile_File_Exists (struct pak_access_s* myPack, const char* storedName);
fbool					PakFile_Is_Edit_Mode(struct pak_access_s* myPack);
int						PakFile_NumChanges	(struct pak_access_s* myPack);

void					PakFile_Compress	(const char* filename);

// Special.  Looks for a TGA that matches the timestamp too.
fbool PakFile_File_TGA_Is_Packed (struct pak_access_s* myPack, const char* storeAsName, int timeStamp);


// Outputs a list of files in directory
text1024_t*	PakFile_Directory_Alloc (struct pak_access_s* myPack, int* count);

// archive_wad.c
struct wad_access_s*	WadFile_Auto_Open2 (const char* myWadFileName);
struct wad_access_s*	WadFile_Auto_Open3 (const char* myWadFileName);

struct wad_access_s*	WadFile_Open		(const char* myWadFileName, int myFileMode);
struct wad_access_s*	WadFile_Close		(struct wad_access_s* myWad);

fbool					WadFile_Delete_File (struct wad_access_s* myWad, const char* storeAsName);
fbool					WadFile_Rename_File (struct wad_access_s* myWad, const char* storeAsName, const char* newName);
fbool					WadFile_Add_Image_As(struct wad_access_s* myWad, const char* diskFileName, const char* storeAsName);
fbool					WadFile_Pack_File_As(struct wad_access_s* myWad, const char* diskFileName, const char* storeAsName);

fbool					WadFile_Wad_Info	(const char* myWadFileName);
void					WadFile_Compress	(const char* filename);

// Supporting functions for convenience
int WadFile_Misc_Is_Name_Special			(const char* textureName);  // 0 = No, return strlen extension if special. (_bump, _glow, _luma, etc.)
fbool WadFile_Misc_Is_Texture_Name_Too_Long (const char* textureName);
fbool WadFile_Is_Edit_Mode					(struct wad_access_s* myWad);

fbool WadFile_Misc_Is_Name_Wadable			(const char* textureName);
fbool WadFile_Misc_Is_Image_Size_Wadable	(int width, int height);
char* WadFile_Misc_Make_Wad_Name			(char* outTextureName, const char* sourceFileName);
fbool WadFile_Maybe_Add_Image_File			(struct wad_access_s* myWad, const char* imageFileName);

