//**********************************************************************************************************\
//
//	environment.h:	Smoothing out platform and compiler variations. 
//
//**********************************************************************************************************/

#ifndef _ENVIRONMENT_H
#define _ENVIRONMENT_H

void* Memory_free (void* ptr);
void* Memory_malloc (const size_t size, const char* Reason);
void* Memory_calloc (const size_t n, const size_t size, const char* Reason);



extern int commandLineArgumentsCount;
extern char **commandLineArguments;
extern char currentWorkingDirectory[MAX_PATH_IS_1024];
char executableAbsoluteName[MAX_PATH_IS_1024];
char executableDirectory[MAX_PATH_IS_1024];


extern int CheckParm (char *check);
void Enter_To_Exit_DebugOnly (void);

// strlcat and strlcpy, from OpenBSD
// Most (all?) BSDs already have them
#if defined(__OpenBSD__) || defined(__NetBSD__) || defined(__FreeBSD__) || defined(MACOSX)
# define HAVE_STRLCAT 1
# define HAVE_STRLCPY 1
#endif

#ifndef HAVE_STRLCAT
/*!
 * Appends src to string dst of size siz (unlike strncat, siz is the
 * full size of dst, not space left).  At most siz-1 characters
 * will be copied.  Always NUL terminates (unless siz <= strlen(dst)).
 * Returns strlen(src) + MIN(siz, strlen(initial dst)).
 * If retval >= siz, truncation occurred.
 */
size_t strlcat(char *dst, const char *src, size_t siz);
#endif  // #ifndef HAVE_STRLCAT

#ifndef HAVE_STRLCPY
/*!
 * Copy src to string dst of size siz.  At most siz-1 characters
 * will be copied.  Always NUL terminates (unless siz == 0).
 * Returns strlen(src); if retval >= siz, truncation occurred.
 */
size_t strlcpy(char *dst, const char *src, size_t siz);

#endif  // #ifndef HAVE_STRLCPY

// Byte ordering

#ifdef __BIG_ENDIAN__

long    LittleLong (long myLong)
#else
#define LittleLong(value) (value)
#endif

// MSVC has a different name for several standard functions

#ifdef _WIN32

#define strcasecmp _stricmp
#define strncasecmp _strnicmp
#define SNPrintf _snprintf
#else
#define	SNPrintf snprintf
#endif

// Not missing, just desired
#define CLAMP(min, x, max)	((x) < (min) ? (min) : (x) > (max) ? (max) : (x))
#define FLOOR max
#define CEILING min

#define roundf(x)   (floorf((x) + 0.5f)) // If we don't have C99 ... this is a bit wrong but whatever

//
// Data Types
//

typedef unsigned char 		byte;


typedef enum
{
	False	= 0,
	True	= 1
}	fbool;

#define Host_FatalError printf
#define Console_Printf printf

void printDateAndTime (void);
fbool System_Arguments_Init (int argc, char** argv);
#endif // !ENVIRONMENT_H