//**********************************************************************************************************\
//
//	image.c:	Image functions
//
//**********************************************************************************************************/

#include "project.h"

#define pIMAGERGBA_TO_PARAMS(imagergba) imagergba->data, imagergba->byteCount, imagergba->pixelcount, imagergba->width, imagergba->height


int Image_Info_AlphaType (const imageRGBA_t* myImage)
{
	return Image_RGBA_GetInfo_AlphaType ( pIMAGERGBA_TO_PARAMS(myImage) );
}

int Image_Info_NumColorsUsed (const imageRGBA_t* myImage)
{
	return Image_RGBA_GetInfo_NumColorsUsed ( pIMAGERGBA_TO_PARAMS(myImage) );
}

fbool Image_Info_IsSizePower2 (const imageRGBA_t* myImage)
{
	if (myImage->width  != RoundUp_To_Power_Of_Two (myImage->width ) ) return False;	// Width isn't power of 2
	if (myImage->height != RoundUp_To_Power_Of_Two (myImage->height) ) return False;	// Height isn't power of 2
	return True;
}

fbool Image_Info_NewSizeFitsImage	(const imageRGBA_t* myImage, int width, int height)
{
	if (myImage->width >= width && myImage->height >= height) return True;
	return False;
}

fbool Image_Info_NewSizeIsSame	(const imageRGBA_t* myImage, int width, int height)
{
	if (myImage->width == width && myImage->height == height) return True;
	return False;
}


void  Image_Info_CenterRectOfSize (const imageRGBA_t* myImage, int width, int height, int* rectCenteredAtX, int* rectCenteredAtY)
{
	*rectCenteredAtX = Center_Position_BigSize_LittleSize_Int (myImage->width,  width); 
	*rectCenteredAtY = Center_Position_BigSize_LittleSize_Int (myImage->height, height);
}


void Image_EnlargeTo_POW2_Realloc	(imageRGBA_t* myImage)
{
	const int POW2width		= RoundUp_To_Power_Of_Two (myImage->width);
	const int POW2height	= RoundUp_To_Power_Of_Two (myImage->height);
	
//	if (POW2width == modifyingImage->width &&  POW2height == modifyingImage->height) 
//		return; // No action required  NO ... screen this out.
	
	Image_Enlarge_Canvas (myImage, POW2width, POW2height, 0);
}


void Image_Do_PasteOperationAt (imageRGBA_t* myImage, pasteop pasteOp, imageRGBA_t* pasteSource, int pasteAtX, int pasteAtY)
{
	Image_RGBA_PasteAt (pasteOp, pasteAtX, pasteAtY, pIMAGERGBA_TO_PARAMS(pasteSource), pIMAGERGBA_TO_PARAMS(myImage) );
}

void Image_Do_PixelOperation (imageRGBA_t* myImage, const pixelop myPixelOperation)
{
	Image_RGBA_Modify(myPixelOperation, pIMAGERGBA_TO_PARAMS(myImage) );
}


void Image_Create_UsingData (imageRGBA_t* myImage, int width, int height, byte* pData)
{
	myImage->width				= width;
	myImage->height				= height;
	myImage->pixelcount			= myImage->width * myImage->height;
	myImage->byteCount			= myImage->pixelcount * RGBA_BytesPerPixel_Is_4;
	myImage->data				= pData;
	
}

void Image_Create_Duplicat		(imageRGBA_t* myImage, imageRGBA_t* duplicateImage)
{
	Image_Create_New			(duplicateImage, myImage->width, myImage->height, 0);
	Image_Do_PasteOperationAt	(duplicateImage, PASTE_REPLACE, myImage, 0, 0);
}

void Image_Destroy_KeepStats_free (imageRGBA_t* myImage)
{
	myImage->data				= Memory_free(myImage->data);	// Goodbye actual data
//	myImage->byteCount			= 0;							// This we do zero out, since this is the memory length ?
	// Keep width, height, pixelcount
}

void Image_Destroy (imageRGBA_t* myImage)
{
	myImage->data				= Memory_free(myImage->data);	// Goodbye actual data
//	myImage->byteCount			= 0;							// This we do zero out, since this is the memory length ?
	// Keep width, height, pixelcount
}


fbool Image_Import_FileX (imageRGBA_t* myImage, const char* loadFileName)
{
	FILE*	fin = fopen (loadFileName, "rb");
	fbool	ReadOK = False;

	if (fin)
	{
		ReadOK = Image_Import_TGA_OpenFile (myImage, fin);
		fclose (fin);
	}

	return ReadOK;
}


// Allocates new of size width, height and destroys old
void Image_Create_New (imageRGBA_t* myImage, int width, int height, unsigned backgroundColor)
{
	Image_Create_UsingData	(myImage, width, height, NULL);

	myImage->data			= Memory_calloc (1, myImage->byteCount, "Enlarge canvas"); // New black image (todo specify bgcolor)
	
	if (backgroundColor)
	{
		unsigned*	colorMe = (unsigned*)myImage->data;
		int			i;
		for (i = 0; i < myImage->pixelcount; i++, colorMe++) // "if" is to avoid wasting time filling with black
			*colorMe = backgroundColor;
	}
}

void Image_Replace (imageRGBA_t* exitingImage, imageRGBA_t* receivingImage)
{
	receivingImage->data		= Memory_free(receivingImage->data);	// Dealloc
	
	receivingImage->width		= exitingImage->width;
	receivingImage->height		= exitingImage->height;
	receivingImage->pixelcount	= exitingImage->pixelcount;
	receivingImage->byteCount	= exitingImage->byteCount;
	receivingImage->data		= exitingImage->data;
}


void Image_Enlarge_Canvas (imageRGBA_t* myImage, int newWidth, int newHeight, unsigned backgroundColor)
{
	imageRGBA_t		enlargedImage;
	imageRGBA_t*	pEnlargedImage = &enlargedImage;

	Image_Create_New			(pEnlargedImage, newWidth, newHeight, backgroundColor);
	Image_Do_PasteOperationAt	(pEnlargedImage, PASTE_REPLACE, myImage, 0, 0);
	Image_Replace				(pEnlargedImage, myImage);
}

void Image_Resample_By_Percent	(imageRGBA_t* myImage, float percent)
{
	int newWidth	= (int)roundf(myImage->width * percent);
	int newHeight	= (int)roundf(myImage->height * percent);

	Image_Resample_To_Size (myImage, newWidth, newHeight);
}

void Image_Resample_To_Size	(imageRGBA_t* myImage, int newWidth, int newHeight)
{
	imageRGBA_t		_resized;
	imageRGBA_t*	resized = &_resized;

	Image_Create_New (resized, newWidth, newHeight, 0);
	
	resized->data = 
	Image_RGBA_Resample	(	myImage->data, myImage->byteCount, myImage->width, myImage->height,
							resized->data, resized->byteCount, resized->width, resized->height);

	Image_Replace		(resized, myImage); 

}


fbool Image_Query_File_BitsPerPixel (const char* fileName)
{
	int myBitsPerPixel;
	Image_Query_File_Dimensions (fileName, NULL, NULL, &myBitsPerPixel);
	return myBitsPerPixel;
}