//**********************************************************************************************************\
//
//	image_convert.c:	Image export or import (wrapped around an imageRGBA_t)
//
//**********************************************************************************************************/

#include "project.h"






fbool  Image_Export_TGA_File_With_Header (const imageRGBA_t* myImage, const char* fileToWrite, const byte* header, unsigned char headerSize)
{
	return Image_RGBA_TGA_Write_File (fileToWrite, myImage->data, myImage->width, myImage->height, header, headerSize);
}

fbool  Image_Export_TGA_File (const imageRGBA_t* myImage, const char* fileToWrite)
{
	return Image_Export_TGA_File_With_Header (myImage, fileToWrite, NULL, 0);
}

fbool Image_Import_TGA_OpenFile (imageRGBA_t* myImage, FILE* tgaFileHandle)
{
	fbool	FileIsOk;
	
	myImage->data = Image_RGBA_TGA_Read_OpenFile_Alloc(tgaFileHandle, NULL, &myImage->width, &myImage->height, &FileIsOk);
	
	if (FileIsOk == False)	return False;

	// data, width and height are filled ... so just need ...
	myImage->pixelcount			= myImage->width * myImage->height;
	myImage->byteCount			= myImage->pixelcount * RGBA_BytesPerPixel_Is_4;
	
	return True; // No problems ...
}

typedef struct
{
	int	width; 
	int	height;
} lmp_t;

fbool Image_Export_LMP_File (const imageRGBA_t* myImage, const char* fileToWrite)
{
	lmp_t	lmpHeader	= {myImage->width, myImage->height};	
	FILE*	fout		= fopen (fileToWrite, "wb");
	
	fbool	writeOK		= False;

	if (fout)
	{
		byte*	QPALImage	= Image_Export_QPAL_Data_Alloc (myImage);
		
		do 
		{
			// fwrite returns the number of full items actually written, which may be less than count if an error occurs
			if ( !File_Safe_Write (fout, &lmpHeader, sizeof(lmpHeader))   ) break;
			if ( !File_Safe_Write (fout, QPALImage, myImage->pixelcount) ) break;
		
			writeOK =  True;
		} while (0);

		QPALImage = Memory_free (QPALImage);
		fclose (fout);
	} 

	return writeOK;
}

fbool Image_Export_Mip_File (const imageRGBA_t* myImage, const char* fileName, const char* textureName)
{
	FILE*	fout	= fopen (fileName, "wb");
	fbool	writeOK = False;

	if (fout)
	{
		writeOK = Image_Export_Mip_OpenFile (myImage, fout, textureName);
		fclose (fout);
	}

	return writeOK;

}



byte* Image_Export_QPAL_Data_Alloc (const imageRGBA_t* myImage)
{
	byte*		outData = Memory_malloc (myImage->pixelcount, "QPAL image");

	{
		unsigned*	inCursor = (unsigned*)myImage->data;
		byte*		outCursor = outData;

		byte		red, green, blue, alpha;
		int			x, y;

		for (y = 0; y < myImage->height; y++)
			for (x = 0; x < myImage->width; x++, inCursor++, outCursor++)
			{
				Pixel_RGBA_To_Components (*inCursor, &red, &green, &blue, &alpha);
				*outCursor = PixelByte_Get_Best_Quake_NoFullbright (red, green, blue);
			}
	}
	return outData;
}

byte* Image_Export_QPAL_With_Alpha255Color_Data_Alloc (const imageRGBA_t* myImage)
{
	byte*		outData = Memory_malloc (myImage->pixelcount, "QPAL image");

	{
		unsigned*	inCursor = (unsigned*)myImage->data;
		byte*		outCursor = outData;

		byte		red, green, blue, alpha;
		int			x, y;

		for (y = 0; y < myImage->height; y++)
			for (x = 0; x < myImage->width; x++, inCursor++, outCursor++)
			{
				Pixel_RGBA_To_Components (*inCursor, &red, &green, &blue, &alpha);
				if (alpha < 127)
					*outCursor = 255; // Quake transparent color
				else
					*outCursor = PixelByte_Get_Best_Quake_NoFullbright (red, green, blue);
			}
	}
	return outData;
}

typedef struct 
{
   byte		name[16];
   unsigned width;
   unsigned height;
   unsigned offset[4];
} miptex_t;				

int sMipSize_For_Width_Height (int width, int height)
{
	int result = 0;
	// Sure could use 85/64 magic number.  8*8 + 4*4 + 2*2 + 1*1 = 64 + 16 + 4 + 1 = 85
	// But that cheat doesn't explain anything as to how it is calculated.
	// And is just one more thing to confuse someone reading your code.
	result += width * height;
	result += (width/2 * height/2);
	result += (width/4 * height/4);
	result += (width/8 * height/8);

	return result;
}

// Exports a blob of memory that contains a MIP
byte* Image_Export_Mip_Data_Alloc  (const imageRGBA_t *myImage, int* setByteLength, const char* textureName)
{
	int			dataStart	= sizeof(miptex_t);
	int			blobsize	= dataStart + sMipSize_For_Width_Height (myImage->width, myImage->height);
	byte*		mipExport	= Memory_calloc (blobsize, 1, "Miptex memory blob");
	miptex_t*	asMipData	= (miptex_t*)mipExport;


	// Fill in the header
	StringLCopy (asMipData->name, textureName);

	asMipData->width		= myImage->width;
	asMipData->height		= myImage->height;
	asMipData->offset[0]	= sizeof(miptex_t);
	asMipData->offset[1]	= asMipData->offset[0] + asMipData->width * asMipData->height;
	asMipData->offset[2]	= asMipData->offset[1] + (asMipData->width / 2) * (asMipData->height / 2);
	asMipData->offset[3]	= asMipData->offset[2] + (asMipData->width / 4) * (asMipData->height / 4);

	do 
	{
		int		width[4]	= { myImage->width,  myImage->width  / 2, myImage->width  / 4, myImage->width  / 8};
		int		height[4]	= { myImage->height, myImage->height / 2, myImage->height / 4, myImage->height / 8};
		byte*	QPALImage[4];
		int		pixcount[4];
		int		mipLevel;

		for (mipLevel = 0; mipLevel < 4; mipLevel ++)
			pixcount[mipLevel] = width[mipLevel] * height[mipLevel];

		QPALImage[0]	= Image_Export_QPAL_Data_Alloc (myImage);
		QPALImage[1]	= Memory_calloc (pixcount[1], 1, "mip2 space");
		QPALImage[2]	= Memory_calloc (pixcount[2], 1, "mip4 space");
		QPALImage[3]	= Memory_calloc (pixcount[3], 1, "mip8 space");
		
		for (mipLevel = 1; mipLevel < 4; mipLevel ++)
		{
			int scale = (int)pow (2, mipLevel); // 2 ^ 1 or 2^2 or 2^3 = 8
			int row, col;
			for (row = 0; row < height[mipLevel]; row ++)
				for (col =0; col < width[mipLevel]; col ++) 
					QPALImage[mipLevel][row * width[mipLevel] + col] = QPALImage[0][ ( row * myImage->width * scale ) + (col * scale) ];
		}

		// Ugly looking .. copy into memory blob
		memcpy (&mipExport[asMipData->offset[0]], QPALImage[0], pixcount[0]); 
		memcpy (&mipExport[asMipData->offset[1]], QPALImage[1], pixcount[1]);
		memcpy (&mipExport[asMipData->offset[2]], QPALImage[2], pixcount[2]);
		memcpy (&mipExport[asMipData->offset[3]], QPALImage[3], pixcount[3]);
	} while (0);

	return mipExport;
}

// myWAD->dataBytesWritten    +=  File_Safe_Write (myWAD->fileHandle, pad,		     sizeof(pad));			// 2
//			myWAD->dataBytesWritten    +=  File_Safe_Write (myWAD->fileHandle, quakePalette, sizeof(quakePalette));	// 768
//			myWAD->dataBytesWritten    +=  File_Safe_Write (myWAD->fileHandle, pad,		     sizeof(pad));			// 2

fbool Image_Export_Mip_OpenFile (const imageRGBA_t* myImage, FILE* fout, const char* textureName)
{
	miptex_t	mipHeader;
	byte*		QPALImage[4];
	int			row, col, scale, mipLevel;

	int			width[4]  = { myImage->width,  myImage->width  / 2, myImage->width  / 4, myImage->width  / 8};
	int			height[4] = { myImage->height, myImage->height / 2, myImage->height / 4, myImage->height / 8};
	int			pixcount[4];

	for (mipLevel = 0; mipLevel < 4; mipLevel ++)
		pixcount[mipLevel] = width[mipLevel] * height[mipLevel];
	
	if (!IS_MULTIPLE_OF_16(myImage->width) || !IS_MULTIPLE_OF_16(myImage->height))
		return False;

	do	// mipmap stage
	{
		if (textureName[0] == '{' )
			QPALImage[0]	= Image_Export_QPAL_With_Alpha255Color_Data_Alloc (myImage);
		else
			QPALImage[0]	= Image_Export_QPAL_Data_Alloc (myImage);
		QPALImage[1]	= Memory_calloc (pixcount[1], 1, "mip2 space");
		QPALImage[2]	= Memory_calloc (pixcount[2], 1, "mip4 space");
		QPALImage[3]	= Memory_calloc (pixcount[3], 1, "mip8 space");
	
		for (mipLevel = 1; mipLevel < 4; mipLevel ++)
		{
			scale = (int)pow (2, mipLevel); // 2 ^ 1 or 2^2 or 2^3 = 8
			for (row = 0; row < height[mipLevel]; row ++)
				for (col =0; col < width[mipLevel]; col ++) 
					QPALImage[mipLevel][row * width[mipLevel] + col] = QPALImage[0][ ( row * myImage->width * scale ) + (col * scale) ];


		}

	} while (0);

	do // Header stage
	{
		
		memset (&mipHeader, 0, sizeof(mipHeader) );

		StringLCopy (mipHeader.name, textureName);

		mipHeader.width		= myImage->width;
		mipHeader.height	= myImage->height;
		mipHeader.offset[0]	= 40;
		mipHeader.offset[1] = mipHeader.offset[0] + pixcount[0];
		mipHeader.offset[2] = mipHeader.offset[1] + pixcount[1];
		mipHeader.offset[3] = mipHeader.offset[2] + pixcount[2];
	} while (0);

	fwrite (&mipHeader, sizeof(mipHeader), 1, fout);
	fwrite (QPALImage[0], pixcount[0], 1, fout); 
	fwrite (QPALImage[1], pixcount[1], 1, fout);
	fwrite (QPALImage[2], pixcount[2], 1, fout);
	fwrite (QPALImage[3], pixcount[3], 1, fout);

	// Clean up
	QPALImage[0] = Memory_free (QPALImage[0]);
	QPALImage[1] = Memory_free (QPALImage[1]);
	QPALImage[2] = Memory_free (QPALImage[2]);
	QPALImage[3] = Memory_free (QPALImage[3]);
	
	return True;
}






