//**********************************************************************************************************\
//
//	pixel_and_color.c:	Pixel manipulation and color stuff
//
//**********************************************************************************************************/

#include "project.h"


/////////////////////////////////////////////////////////////////////////////////
//// Bare minimum to use this file:
//
//#include "eX_math_geometry_2d.h"	// Rectarea
//#include <string.h>					// memcpy
//
//#include "eX_image.h"				// To know what a palette is
//#include "eX_image_pixel.h"

void Pixel_Apply_Best_Quake_NoFullbright (byte* red, byte* green, byte *blue);




//// Pixel compoents operations.

void PixelComponent_Swap (byte* component1, byte* component2)
{
	byte extra = *component1;
	*component1 = *component2;
	*component2 = extra; 
}

byte PixelComponent_Get_Average (byte* red, byte* green, byte *blue)
{
	unsigned average	= ((unsigned)*red + *green + *blue) / 3;	// Average the colors
	average				= CLAMP (0, average, 255);		// Clamp to range just in case.  Shouldn't happen?
	return average;
}

int Palette_AlphaInfo_From_GreyInfo (fbool foundGrey, fbool foundWhite, fbool foundBlack)
{
	if (foundGrey)					return COLORINFO_ALPHA_CHANNEL;
	if (foundBlack && foundWhite)	return COLORINFO_ALPHA_ALPHA_MASK;	
	if (foundWhite)					return COLORINFO_ALPHA_SOLID_WHITE;
	
	return COLORINFO_ALPHA_SOLID_BLACK;
}

#define ALPHA_FROM_PIXEL(pixel) ((pixel >> 24) & 0xFF)	// Converts unsigned to alpha 0 - 255 byte
#define PIXEL_STRIP_ALPHA(pixel) (pixel & 0x00FFFFFF)	// Removes the bytes representing alpha

///////////////////////////////////////////////////////////////////////////////
// Pixel RGBA creation FROM (look at data and build pixel)
///////////////////////////////////////////////////////////////////////////////


unsigned Pixel_RGBA_From_Components (byte red, byte green, byte blue, byte alpha)
{
	return ((unsigned)red + ((unsigned)green << 8) + ((unsigned)blue << 16) + ((unsigned)alpha << 24));
}

unsigned Pixel_RGBA_From_Components_Percents (float redPercent, float greenPercent, float bluePercent, float alphaPercent)
{
	byte red	= (byte)(redPercent   * 255.0f);
	byte green	= (byte)(greenPercent * 255.0f);
	byte blue	= (byte)(bluePercent  * 255.0f);
	byte alpha	= (byte)(alphaPercent * 255.0f);
	
	return Pixel_RGBA_From_Components (red, green, blue, alpha);
}

///////////////////////////////////////////////////////////////////////////////
// Pixel RGBA out TO creation (Look at pixel and make some other data)
///////////////////////////////////////////////////////////////////////////////

void Pixel_RGBA_To_Components (unsigned pixel, byte* red, byte* green, byte *blue, byte* alpha)
{
	*red				= (pixel      )  & 0xFF;
	*green				= (pixel >> 8 )  & 0xFF;
	*blue				= (pixel >> 16)  & 0xFF;
	*alpha				= (pixel >> 24)  & 0xFF;	
}

void Pixel_RGBA_To_Components_Percents (unsigned pixel, float* redPercent, float* greenPercent, float* bluePercent, float* alphaPercent)
{
	byte red, green, blue, alpha;	
	Pixel_RGBA_To_Components (pixel, &red, &green, &blue, &alpha);
	
	*redPercent				= red   / 255.0f;
	*greenPercent			= green / 255.0f;
	*bluePercent			= blue  / 255.0f;
	*alphaPercent			= alpha / 255.0f;
}




///////////////////////////////////////////////////////////////////////////////
// Pixel RGBA out TO creation where this is general information
///////////////////////////////////////////////////////////////////////////////




void Pixel_RGBA_To_AlphaInfo (unsigned pixel, fbool* foundGrey, fbool* foundWhite, fbool* foundBlack)
{
	const unsigned pixelalpha	= ALPHA_FROM_PIXEL(pixel); // (pixel >> 24) & 0xFF; // 0xFF = 256;
	
	if (0 < pixelalpha       && pixelalpha  < 255)	*foundGrey  = True;	
	if (*foundBlack == False && pixelalpha ==   0)	*foundBlack = True;
	if (*foundWhite == False && pixelalpha == 255)	*foundWhite = True;	
	
}

void Pixel_RGBA_CheckAdd_To_Palette (unsigned pixel, palette_t* palette)
{
	unsigned colorForPixel = PIXEL_STRIP_ALPHA(pixel); // Strip alpha
	int findColorIndex;

	// Find currentPixel in colorsTable	
	for (findColorIndex = 0; findColorIndex < palette->numColors; findColorIndex ++)
		if (palette->colors[findColorIndex] == colorForPixel) return; // Found, nothing to do ...
	
	// If color was not found, add it the end of the table and increase numColorsTable
	palette->colors[palette->numColors++] = colorForPixel;
}


unsigned Pixel_RGBA_PixelOperation (unsigned myPixel, pixelop pixelOperation)
{
	byte red					= (myPixel      )  & 0xFF;
	byte green					= (myPixel >> 8 )  & 0xFF;
	byte blue					= (myPixel >> 16)  & 0xFF;
	byte alpha					= (myPixel >> 24)  & 0xFF;
	
	switch (pixelOperation)
	{
		case PIXOP_GREYSCALE:			red = green = blue = PixelComponent_Get_Average (&red, &green, &blue);			break;				
		case PIXOP_NEGATIVE_IMAGE:		red	= ~red; green = ~green; blue = ~ blue; /* ~ is bitwise NOT */		break;
		case PIXOP_SWAP_RED_GREEN:		PixelComponent_Swap (&red, &green);										break;
		case PIXOP_SWAP_GREEN_BLUE:		PixelComponent_Swap (&green, &blue);									break;
		case PIXOP_SWAP_BLUE_RED:		PixelComponent_Swap (&blue, &red);										break;
		case PIXOP_ALPHA_FROM_COLOR:	alpha = PixelComponent_Get_Average (&red, &green, &blue);				break;
		case PIXOP_REMOVE_RED:			red = 0;																break;
		case PIXOP_REMOVE_GREEN:		green = 0;																break;
		case PIXOP_REMOVE_BLUE:			blue = 0;																break;
			
			
		case PIXOP_QUAKE_COLORS_NO_FULLBRIGHT:	Pixel_Apply_Best_Quake_NoFullbright (&red, &green, &blue);		break;
		default:																								break;
	}
	
	return Pixel_RGBA_From_Components (red, green, blue, alpha); 	
}

unsigned Pixel_RGBA_PasteOp_Blend (pasteop pasteOp, unsigned sourcePixel, unsigned destPixel)
{
	return sourcePixel; // For now!
}

#include "quake_palette.h"
void Pixel_Apply_Best_Palette_Index (byte* red, byte* green, byte *blue, byte* myPalette, int* myPaletteNumColors);


void Pixel_Apply_Best_Quake_NoFullbright (byte* red, byte* green, byte *blue)
{
	int numNoFullBrightColors = 224;
	Pixel_Apply_Best_Palette_Index (red, green, blue, quakePalette, &numNoFullBrightColors);
}

void Pixel_Apply_Best_Palette_Index (byte* red, byte* green, byte *blue, byte* myPalette, int* myPaletteNumColors)
{
	int				bestColorIndex		=  -1;
	unsigned		bestColorDistance	=  0-1; // Should be a big number
	const byte*		currentPaletteIndex	= myPalette;
	int				i;

	for (i = 0; i < *myPaletteNumColors; i++, currentPaletteIndex += 3 )
	{
		const int		redDistance		= *red   - currentPaletteIndex[0];
		const int		greenDistance	= *green - currentPaletteIndex[1];
		const int		blueDistance	= *blue  - currentPaletteIndex[2];
		const unsigned	ColorDistance	= SQUARED(redDistance) + SQUARED(greenDistance) + SQUARED(blueDistance); // Could sqrt this but no reason to do so
		
		if (ColorDistance < bestColorDistance)
		{
			bestColorDistance	= ColorDistance;
			bestColorIndex		= i;
		}
	}
	
	*red	= myPalette[bestColorIndex * 3 + 0];
	*green	= myPalette[bestColorIndex * 3 + 1];
	*blue	= myPalette[bestColorIndex * 3 + 2];
	
}



byte PixelByte_Get_Best_Palette_Index (byte red, byte green, byte blue, byte* myPalette, int* myPaletteNumColors)
{
	int				bestColorIndex		=  -1;
	unsigned		bestColorDistance	=  0-1; // Should be a big number
	const byte*		currentPaletteIndex	= myPalette;
	int				i;

	for (i = 0; i < *myPaletteNumColors; i++, currentPaletteIndex += 3 )
	{
		const int		redDistance		= red   - currentPaletteIndex[0];
		const int		greenDistance	= green - currentPaletteIndex[1];
		const int		blueDistance	= blue  - currentPaletteIndex[2];
		const unsigned	ColorDistance	= SQUARED(redDistance) + SQUARED(greenDistance) + SQUARED(blueDistance); // Could sqrt this but no reason to do so
		
		if (ColorDistance < bestColorDistance)
		{
			bestColorDistance	= ColorDistance;
			bestColorIndex		= i;
		}
	}
	
	return (byte)bestColorIndex;
	
}

byte PixelByte_Get_Best_Quake_NoFullbright (byte red, byte green, byte blue)
{
	static int numNoFullBrightColors = 224;
	return PixelByte_Get_Best_Palette_Index (red, green, blue, quakePalette, &numNoFullBrightColors);
}