//**********************************************************************************************************\
//
//	image_rgba.c:	Image functions
//
//**********************************************************************************************************/



#include "project.h"

/////////////////////////////////////////////////////////////////////////////////
//// Bare minimum to use this file:
//
////#include "eX_math_geometry_2d.h"	// Rectarea
//#include "eX_math_screenrects.h"	// Rect functions
//#include <string.h>					// memcpy
//
//#include "eX_image_pixel.h"			// Access to pixel manipulation functions





///////////////////////////////////////////////////////////////////////////////
// Elementary functions
// Rules for now: Must take byte, buffersize, pixelcount, width, height
// Must be rather easy to understand what is going on.
///////////////////////////////////////////////////////////////////////////////



// Weakness cannot take range.  Then again, cannot take selection.  Gotta start somewhere.
int Image_RGBA_GetInfo_NumColorsUsed(const byte *source, size_t sourceSize, int pixelCount, int sourceWidth, int sourceHeight)
{

	unsigned*	colorsTableData = Memory_malloc (pixelCount, "Maximum possible colors");
//	palette_t	myPalette = {sizeof(colorsTableData)/sizeof(colorsTableData[0]), colorsTableData, 0}; 
	palette_t	myPalette = {sizeof(colorsTableData)/sizeof(colorsTableData[0]), colorsTableData, 0}; 
	unsigned*	thisPixel = (unsigned*) source;
	int			i;

	for (i = 0 ; i < pixelCount /* <-- that is correct */ ; i++, thisPixel ++)
		Pixel_RGBA_CheckAdd_To_Palette (*thisPixel, &myPalette);
	
	colorsTableData = Memory_free (colorsTableData);
	return myPalette.numColors;
}


// Weakness cannot take range.  Then again, cannot take selection.  Gotta start somewhere.
int Image_RGBA_GetInfo_AlphaType	(const byte *buffer, size_t buffersize, int pixelCount, int width, int height)
{
	fbool		foundBlack		= False;
	fbool		foundWhite		= False;
	fbool		foundGrey		= False;
	
	unsigned*	thisPixel		= (unsigned*) buffer;
	int			i;

	for (i = 0 ; i < pixelCount && foundGrey == False ; i++, thisPixel ++)
		Pixel_RGBA_To_AlphaInfo (*thisPixel, &foundGrey, &foundWhite, &foundBlack);
	
	return Palette_AlphaInfo_From_GreyInfo (foundGrey, foundWhite, foundBlack);
}



byte* Image_RGBA_Modify_RectArea (pixelop pixelOperation, int x1, int y1, int x2, int y2, byte *buffer, size_t bufferSize, int pixelCount, int width, int height)
{
	unsigned*	imagePix		= (unsigned*)buffer;
	int			row, col;

	for (row = y1; row <= y2; row ++)
		for (col = x1; col <= x2; col ++)
		{
			unsigned* thisPixel = &imagePix[row  * width + col];   //			unsigned* thisPixel = &buffer[row  * sourceDestWidth + col];
			*thisPixel = Pixel_RGBA_PixelOperation (*thisPixel, pixelOperation);
		}
	
	return buffer;
}

byte* Image_RGBA_Modify (pixelop pixelOperation, byte *buffer, size_t bufferSize, int pixelCount, int width, int height)
{	
	return Image_RGBA_Modify_RectArea (pixelOperation,  0, 0, width - 1, height - 1, buffer, bufferSize, pixelCount, width, height);
}

#if 0
byte* Image_RGBA_Transform (transform_simple_op transformOperation, byte *buffer, size_t bufferSize, int pixelCount, int* width, int* height)
{	
	const fbool swapWidthHeight		= (transformOperation == TRANSFORM_ROTATE_CLOCKWISE90 || transformOperation == TRANSFORM_ROTATE_COUNTERCLOCKWISE90);
	
	unsigned*	oldImage			= (unsigned*)buffer;
	const int	oldWidth			= *width;
	const int	oldHeight			= *height;
	
	unsigned*	newImage			= Memory_malloc(bufferSize, "New Buffer");
	int			newWidth			= oldWidth;
	int			newHeight			= oldHeight;
	
	unsigned*	newPixel			= newImage;

	if (swapWidthHeight)
	{
		newWidth	= oldHeight;
		newHeight	= oldWidth;
	}
	
	for (int row = 0; row < oldHeight; row ++)
		for (int col = 0; col < oldWidth; col ++, newPixel ++)
		{
			int oldrow, oldcol;
			switch (transformOperation)
			{
				case TRANSFORM_ROTATE_COUNTERCLOCKWISE90:
					oldrow = col;					oldcol = (oldHeight - 1) - row;	break;
				case TRANSFORM_ROTATE_CLOCKWISE90:	oldrow = (oldWidth - 1) - col;	oldcol = row;					break;
				case TRANSFORM_FLIP_VERTICAL:		oldrow = (oldHeight - 1) - row;	oldcol = col;					break;
				case TRANSFORM_FLIP_HORIZONTAL:		oldrow = row;					oldcol = (oldWidth - 1) - col;	break; 
				case TRANSFORM_ROTATE_180:			oldrow = (oldHeight - 1) - row;	oldcol = (oldWidth - 1) - col;	break;
				default:																							break;
			}
			*newPixel = oldImage[oldrow * oldWidth + oldcol];
		}
	
	
	// Destroy old image
	buffer = Memory_free (buffer);
	
	*width	= newWidth;
	*height	= newHeight;
	return (byte*) newImage;
}
#endif

byte* Image_RGBA_Transform (transform_simple_op transformOperation, byte *buffer, size_t bufferSize, int pixelCount, int* width, int* height)
{	
	const fbool swapWidthHeight		= (transformOperation == TRANSFORM_ROTATE_CLOCKWISE90 || transformOperation == TRANSFORM_ROTATE_COUNTERCLOCKWISE90);
	
	unsigned*	oldImage			= (unsigned*)buffer;
	const int	oldWidth			= *width;
	const int	oldHeight			= *height;
	
	unsigned*	newImage			= Memory_malloc(bufferSize, "New Buffer");
	int			newWidth			= oldWidth;
	int			newHeight			= oldHeight;
	
	unsigned*	newPixel			= newImage;
	
	int			row, col;

	if (swapWidthHeight)
	{
		newWidth	= oldHeight;
		newHeight	= oldWidth;
	}
	
	for (row = 0; row < newHeight; row ++)
		for (col = 0; col < newWidth; col ++, newPixel ++)
		{
			int oldrow = -99999, oldcol = - 99999;
			switch (transformOperation)
			{
		case TRANSFORM_ROTATE_COUNTERCLOCKWISE90:	oldrow = col;					oldcol = (oldWidth - 1) - row;	break;
				case TRANSFORM_ROTATE_CLOCKWISE90:	oldrow = (oldHeight - 1) - col;	oldcol = row;					break;

				case TRANSFORM_FLIP_VERTICAL:		oldrow = (oldHeight - 1) - row;	oldcol = col;					break;
				case TRANSFORM_FLIP_HORIZONTAL:		oldrow = row;					oldcol = (oldWidth - 1) - col;	break; 
				case TRANSFORM_ROTATE_180:			oldrow = (oldHeight - 1) - row;	oldcol = (oldWidth - 1) - col;	break;
				default:	printf ("Unknown transformation"); 																						
							break;
			}
			*newPixel = oldImage[oldrow * oldWidth + oldcol];
		}
	
	
	// Destroy old image
	buffer = Memory_free (buffer);
	
	*width	= newWidth;
	*height	= newHeight;
	return (byte*) newImage;
}






byte* Image_RGBA_PasteAt (pasteop pasteOp, int x, int y, const byte* pasteSource, size_t pasteSourceSize, int pasteSourcePixelCount, int pasteSourceWidth, int pasteSourceHeight, byte* recipient, size_t recipientSize, int recipientPixelCount, int recipientWidth, int recipientHeight)
{
	// In an unclipped world, this is the answer

	ScreenXY    pasteOffset			=   {x, y};
	ScreenRect	recipientBoxReal    = { {0, 0} , {recipientWidth - 1,   recipientHeight - 1   } };
//	ScreenRect  pasteSourceBoxReal  = { {0, 0} , {pasteSourceWidth - 1, pasteSourceHeight - 1 } };	// For reference I guess since we do not use (but should!)
	ScreenRect	pasteBoxOffset      = { {x, y} , {x + pasteSourceWidth - 1, y + pasteSourceHeight - 1} }; 

	if ( ScreenRect_DoOverlap (recipientBoxReal, pasteBoxOffset) )
	{
		ScreenRect	recipientOverlapBox = SCREENRECT_OF_OVERLAP (recipientBoxReal, pasteBoxOffset);
		unsigned* recipientRead		= (unsigned*)recipient;
		unsigned* pasteSourceRead	= (unsigned*)pasteSource;
		int			recipientRow, recipientCol;
		for (recipientRow = recipientOverlapBox.topLeft.y; recipientRow <= recipientOverlapBox.bottomRight.y; recipientRow ++)
			for (recipientCol = recipientOverlapBox.topLeft.x; recipientCol <= recipientOverlapBox.bottomRight.x; recipientCol ++)
			{
				// Adjust read position to the offset
				unsigned* recipientWrite		= &recipientRead[recipientRow * recipientWidth + recipientCol];
				unsigned  pixelFromPaste		= pasteSourceRead[(recipientRow + pasteOffset.y) * pasteSourceWidth + (recipientCol + pasteOffset.x)];
				unsigned  pixelFromRecipient	= *recipientWrite;
				// Write it to the paste recipient
				*recipientWrite = Pixel_RGBA_PasteOp_Blend (pasteOp, pixelFromPaste, pixelFromRecipient); // source, dest 
			}

	}
	
	return recipient;
}



///////////////////////////////////////////////////////////////////////////////
//
//	Image loading functions.  TGA only at the moment.  Will likely include
//	texture formats that do not require a lib or dll to load.  PCX ?
//
//	This means no PNG or JPG.  Support for those requires libraries and on 
//	some platforms it may be a pain to get a library or worse, there cannot
//	be a true library.
//
///////////////////////////////////////////////////////////////////////////////

// Image functions that don't require anything
// It would be great to optimize this to just check the alpha byte.
// But to keep this endian neutral ...
// Impossible to tell what could happen in a few years, stranger things have
// happened.


//=========================================================
//
// Targa
//
//=========================================================

#define	IMAGE_MAX_DIMENSIONS	4096
#define TGA_MAXCOLORS 16384

/* Definitions for image types. */
#define TGA_Null	0	/* no image data */
#define TGA_Map		1	/* Uncompressed, color-mapped images. */
#define TGA_RGB		2	/* Uncompressed, RGB images. */
#define TGA_Mono	3	/* Uncompressed, black and white images. */
#define TGA_RLEMap	9	/* Runlength encoded color-mapped images. */
#define TGA_RLERGB	10	/* Runlength encoded RGB images. */
#define TGA_RLEMono	11	/* Compressed, black and white images. */
#define TGA_CompMap	32	/* Compressed color-mapped data, using Huffman, Delta, and runlength encoding. */
#define TGA_CompMap4	33	/* Compressed color-mapped data, using Huffman, Delta, and runlength encoding. 4-pass quadtree-type process. */

/* Definitions for interleave flag. */
#define TGA_IL_None	0	/* non-interleaved. */
#define TGA_IL_Two	1	/* two-way (even/odd) interleaving */
#define TGA_IL_Four	2	/* four way interleaving */
#define TGA_IL_Reserved	3	/* reserved */

/* Definitions for origin flag */
#define TGA_O_UPPER	0	/* Origin in lower left-hand corner. */
#define TGA_O_LOWER	1	/* Origin in upper left-hand corner. */

typedef struct _TargaHeader {
	unsigned char 	id_length, colormap_type, image_type;
	unsigned short	colormap_index, colormap_length;
	unsigned char	colormap_size;
	unsigned short	x_origin, y_origin, width, height;
	unsigned char	pixel_size, attributes;
} TargaHeader;

static int fgetLittleShort (FILE *f);

static int fgetLittleShort (FILE *f)
{
	byte	b1, b2;
	
	b1 = fgetc(f);
	b2 = fgetc(f);
	
	return (short)(b1 + b2*256);
}


#if 0
/*
 =============
 Image_WriteTGA
 =============
 */
fbool Image_WriteTGA_ToOpenFile_RGBA (FILE *fout, const char* filename_nopath, const byte *pixels, int width, int height, int BitsPerPixel, fbool IsUpsideDown)
{
#define TGA_HEADER_SIZE 18	
	fbool		retval			= True;
	int			size			= width * height * ((BitsPerPixel == 32) ? 4 : 3);
	int			buffersize_tga	= size + TGA_HEADER_SIZE;  // pixel data plus the header
	byte*		buffer			= Memory_malloc (buffersize_tga, "Temp write TGA");	// Allocate memory
	
	// Header
	memset (buffer, 0, TGA_HEADER_SIZE);
	
	buffer[2]		= 2;
	buffer[12]		= width & 255;
	buffer[13]		= width >> 8;
	buffer[14]		= height & 255;
	buffer[15]		= height >> 8;
	buffer[16]		= BitsPerPixel;  // 32 or 24;
	
	if (IsUpsideDown)
		buffer[17]	= 0x20; //upside-down attribute
	
	// Pixel data
	memcpy (buffer + TGA_HEADER_SIZE, pixels, size);
	{
		byte	temp;
		int		bytes = BitsPerPixel / 8;
		int		i;

		for (i = TGA_HEADER_SIZE ; i < buffersize_tga /* = size + TGA_HEADER_SIZE*/ ; i += bytes)
		{
			temp = buffer[i];
			
			// Swap red and blue bytes
			buffer[i + 0] = buffer[i + 2];
			buffer[i + 2] = temp;
		}
	}
	
	fwrite (buffer, 1, buffersize_tga, fout);
	
fail:
	buffer = Memory_free (buffer);
	
	return retval; // I guess?
}
#endif



fbool Image_RGBA_TGA_Write_ToOpenFile (FILE *fout, const char* filename_nopath, const byte *pixels, int width, int height, int BitsPerPixel, fbool IsUpsideDown, const byte* header, unsigned char headerSize)
{
#define TGA_HEADER_SIZE 18	
	fbool		retval			= True;
	int			size			= width * height * ((BitsPerPixel == 32) ? 4 : 3);
	int			true_headersize	= headerSize + TGA_HEADER_SIZE;
	int			buffersize_tga	= size + true_headersize;  // pixel data plus the header
	byte*		buffer			= Memory_malloc (buffersize_tga, "Temp write TGA");	// Allocate memory
	
	// Header
	memset (buffer, 0, TGA_HEADER_SIZE); // Size of the TGA Header at 0.  We aren't using one
	buffer[0]		= headerSize;
	buffer[2]		= 2;				// Image Type Code 
	buffer[12]		= width & 255;		// Width height bit
	buffer[13]		= width >> 8;		// Width low bit
	buffer[14]		= height & 255;		// Height low bit
	buffer[15]		= height >> 8;		// Height high bit
	buffer[16]		= BitsPerPixel;		// Bits Per Pixel 32 or 24
	
	if (IsUpsideDown)
		buffer[17]	= 0x20;				// Image Descriptor Byte: upside-down attribute
	
	if (header)
		memcpy (&buffer[18], header, headerSize);

//	// Pixel data
//	memcpy (buffer + TGA_HEADER_SIZE, pixels, size);

	{
		int		srcBytesPerPixel	= RGBA_BytesPerPixel_Is_4;	// Because this is an RGBA function, so it is always 4.
		int		destBytesPerPixel	= BitsPerPixel / 8;			// Could be 3 or 4

		size_t	cursor_source		= 0;						// Start 0 bytes into pixel buffer
		size_t	cursor_dest			= true_headersize;			// Start 18 bytes into memory buffer (plus id length)

		int		pixelsCount			= width * height;
		
		int		i;

		for (i = 0; i < pixelsCount; i ++, cursor_source += srcBytesPerPixel, cursor_dest += destBytesPerPixel)
		{
			buffer[cursor_dest + 0] = pixels[cursor_source + 2];		// Swap red and blue
			buffer[cursor_dest + 1] = pixels[cursor_source + 1];		// Green is green
			buffer[cursor_dest + 2] = pixels[cursor_source + 0];		// Swap red and blue

			if (destBytesPerPixel == RGBA_BytesPerPixel_Is_4)
				buffer[cursor_dest + 3] = pixels[cursor_source + 3];	// Copy alpha if we are writing that
		}
	}
	
	fwrite (buffer, 1, buffersize_tga, fout);
	
fail:
	buffer = Memory_free (buffer);
	
	return retval; // I guess?
}

fbool Image_RGBA_TGA_Write_File (const char* filename, const byte* imageData, int width, int height, const byte* header, unsigned char headerSize)
{
	FILE*	fout = fopen (filename, "wb");
	fbool	didSucceed = False;

	if (fout)
	{
		fbool	hasAlpha = COLORINFO_ALPHA_SOLID_WHITE != Image_RGBA_GetInfo_AlphaType (imageData, width * height * RGBA_BytesPerPixel_Is_4, width * height, width, height);
		int		bitsPerPixelsToWrite = hasAlpha ? 32 : 24;

		didSucceed = Image_RGBA_TGA_Write_ToOpenFile (fout, NULL, imageData, width, height, bitsPerPixelsToWrite, True, header, headerSize);
		fclose (fout);
	}

	return didSucceed;
}

// Cursor needs to be at the start of the "file"
int Image_RGBA_TGA_Read_TimeStamp_OpenFile (FILE *fin)
{
	char	marker[3] = {0,0,0};
	int	secondsSince1970;
	TargaHeader	header;

	fread (&header, TGA_HEADER_SIZE, 1, fin);

	if (header.id_length !=7)
		return 0;
	
	fread (&marker, 1, 2, fin);

	if (StringStartsWith(marker, "TM") == False) 
		return 0;

	fread (&secondsSince1970, 1, sizeof(secondsSince1970), fin); // sizeof can sort of be relied on here since isn't a struct.

	return (secondsSince1970);
}

fbool Image_RGBA_TGA_Read_WidthHeight_OpenFile (FILE *fin, int* setWidth, int* setHeight, int* setBitsPerPixel)
{
	TargaHeader	header;
	memset (&header, 0, sizeof(header) );

	if (!fin)
		return False;

//	fread (&header, sizeof(header), 1, fin);

	header.id_length		= fgetc (fin);
	header.colormap_type	= fgetc (fin);
	header.image_type		= fgetc (fin);
	
	header.colormap_index	= fgetLittleShort (fin);
	header.colormap_length	= fgetLittleShort (fin);
	header.colormap_size	= fgetc (fin);
	header.x_origin			= fgetLittleShort (fin);
	header.y_origin			= fgetLittleShort (fin);
	header.width			= fgetLittleShort (fin);
	header.height			= fgetLittleShort (fin);
	header.pixel_size		= fgetc (fin);

	if (setWidth)		 *setWidth = header.width;	
	if (setHeight)		 *setHeight			= header.height;
	if (setBitsPerPixel) *setBitsPerPixel	    = header.pixel_size;

	if (header.width && header.height) 
		return True; // I guess we succeeded
	
	return False; // 0 width or height
}

fbool Image_Query_File_Dimensions (const char* fileName, int* setWidth, int* setHeight, int* setBitsPerPixel)
{
	FILE* fin = fopen (fileName, "rb");
	fbool	readOK = False;

	if (fin)
	{
		readOK = Image_RGBA_TGA_Read_WidthHeight_OpenFile (fin, setWidth, setHeight, setBitsPerPixel);
		fclose (fin);
	}

	return readOK;
}

//================================
// Image_LoadTGA_RGBA_FromOpenFile
//================================

byte* Image_RGBA_TGA_Read_OpenFile_Alloc (FILE *fin, const char* filename_nopath, int* my_image_width, int* my_image_height, fbool* didSucceed)
{
	char		errhelp[1024];
	int			w, h, x, y, realrow, truerow, baserow, i, temp1, temp2, pixel_size, map_idx;
	int			RLE_count, RLE_flag, size, interleave, origin;
	fbool		mapped, rlencoded;
	byte		*data= NULL, *dst, r, g, b, a, j, k, l, *ColorMap;
	TargaHeader	header;
	
	SNPrintf (errhelp, sizeof(errhelp), "Image_LoadTGA_RGBA_FromOpenFile (%s)", filename_nopath); 
	
	/* load file */
	if (!fin)
		return NULL;
	
	header.id_length		= fgetc (fin);
	header.colormap_type	= fgetc (fin);
	header.image_type		= fgetc (fin);
	
	header.colormap_index	= fgetLittleShort (fin);
	header.colormap_length	= fgetLittleShort (fin);
	header.colormap_size	= fgetc (fin);
	header.x_origin			= fgetLittleShort (fin);
	header.y_origin			= fgetLittleShort (fin);
	header.width			= fgetLittleShort (fin);
	header.height			= fgetLittleShort (fin);
	header.pixel_size		= fgetc (fin);
	header.attributes		= fgetc (fin);
	
	if (header.width > IMAGE_MAX_DIMENSIONS || header.height > IMAGE_MAX_DIMENSIONS)
	{
		Console_Printf ("%s: TGA image (%i x %i) exceeds maximum supported dimensions %i\n", errhelp, (int)header.width, (int)header.height, IMAGE_MAX_DIMENSIONS);
		return NULL;
	}
	
	if (header.id_length != 0)
		fseek (fin, header.id_length, SEEK_CUR);
	
	/* validate TGA type */
	switch (header.image_type)
	{
		case TGA_Map:
		case TGA_RGB:
		case TGA_Mono:
		case TGA_RLEMap:
		case TGA_RLERGB:
		case TGA_RLEMono:
			break;
			
		default:
			Console_Printf ("%s: Unsupported TGA image: Only type 1 (map), 2 (RGB), 3 (mono), 9 (RLEmap), 10 (RLERGB), 11 (RLEmono) TGA images supported\n", errhelp);
			return NULL;
	}
	
	/* validate color depth */
	switch (header.pixel_size)
	{
		case 8:
		case 15:
		case 16:
		case 24:
		case 32:
			break;
			
		default:
			Console_Printf ("%s: Unsupported TGA image: Only 8, 15, 16, 24 or 32 bit images (with colormaps) supported\n", errhelp);
			return NULL;
	}
	
	r = g = b = a = l = 0;
	
	/* if required, read the color map information. */
	ColorMap = NULL;
	mapped = (header.image_type == TGA_Map || header.image_type == TGA_RLEMap) && header.colormap_type == 1;
	if (mapped)
	{
		/* validate colormap size */
		switch (header.colormap_size)
		{
			case 8:
			case 15:
			case 16:
			case 32:
			case 24:
				break;
				
			default:
				Console_Printf ("%s: Unsupported TGA image %s: Only 8, 15, 16, 24 or 32 bit colormaps supported\n", errhelp);
				return NULL;
		}
		
		temp1 = header.colormap_index;
		temp2 = header.colormap_length;
		if ((temp1 + temp2 + 1) >= TGA_MAXCOLORS)
		{
			Console_Printf ("%s: Unsupported TGA image %s: Only 8, 15, 16, 24 or 32 bit colormaps supported\n", errhelp);
			return NULL;
		}
		ColorMap = Memory_malloc (TGA_MAXCOLORS * 4, "temp TGA color map");
		map_idx = 0;
		for (i = temp1 ; i < temp1 + temp2 ; ++i, map_idx += 4)
		{
			/* read appropriate number of bytes, break into rgb & put in map. */
			switch (header.colormap_size)
			{
				case 8:	/* grey scale, read and triplicate. */
					r = g = b = getc (fin);
					a = 255;
					break;
					
				case 15:	/* 5 bits each of red green and blue. */
					/* watch byte order. */
					j = getc (fin);
					k = getc (fin);
					l = ((unsigned int)k << 8) + j;
					r = (byte)(((k & 0x7C) >> 2) << 3);
					g = (byte)((((k & 0x03) << 3) + ((j & 0xE0) >> 5)) << 3);
					b = (byte)((j & 0x1F) << 3);
					a = 255;
					break;
					
				case 16:	/* 5 bits each of red green and blue, 1 alpha bit. */
					/* watch byte order. */
					j = getc (fin);
					k = getc (fin);
					l = ((unsigned int)k << 8) + j;
					r = (byte)(((k & 0x7C) >> 2) << 3);
					g = (byte)((((k & 0x03) << 3) + ((j & 0xE0) >> 5)) << 3);
					b = (byte)((j & 0x1F) << 3);
					a = (k & 0x80) ? 255 : 0;
					break;
					
				case 24:	/* 8 bits each of blue, green and red. */
					b = getc (fin);
					g = getc (fin);
					r = getc (fin);
					a = 255;
					l = 0;
					break;
					
				case 32:	/* 8 bits each of blue, green, red and alpha. */
					b = getc (fin);
					g = getc (fin);
					r = getc (fin);
					a = getc (fin);
					l = 0;
					break;
			}
			ColorMap[map_idx+0] = r;
			ColorMap[map_idx+1] = g;
			ColorMap[map_idx+2] = b;
			ColorMap[map_idx+3] = a;
		}
	}
	
	/* check run-length encoding. */
	rlencoded = (header.image_type == TGA_RLEMap || header.image_type == TGA_RLERGB || header.image_type == TGA_RLEMono);
	RLE_count = RLE_flag = 0;
	
	*my_image_width		= w = header.width;
	*my_image_height	= h = header.height;
	
	size = w * h * 4;
	data = Memory_malloc (size * 1, "TGA loader");
	
	/* read the Targa file body and convert to portable format. */
	pixel_size = header.pixel_size;
	origin = (header.attributes & 0x20) >> 5;
	interleave = (header.attributes & 0xC0) >> 6;
	truerow = baserow = 0;
	for (y=0 ; y<h ; y++)
	{
		realrow = truerow;
		if (origin == TGA_O_UPPER)
			realrow = h - realrow - 1;
		
		dst = data + realrow * w * 4;
		
		for (x=0 ; x<w ; x++)
		{
			/* check if run length encoded. */
			if (rlencoded)
			{
				if (!RLE_count)
				{
					/* have to restart run. */
					i = getc (fin);
					RLE_flag = (i & 0x80);
					if (!RLE_flag)	// stream of unencoded pixels
						RLE_count = i + 1;
					else		// single pixel replicated
						RLE_count = i - 127;
					/* decrement count & get pixel. */
					--RLE_count;
				}
				else
				{
					/* have already read count & (at least) first pixel. */
					--RLE_count;
					if (RLE_flag)
					/* replicated pixels. */
						goto PixEncode;
				}
			}
			
			/* read appropriate number of bytes, break into RGB. */
			switch (pixel_size)
			{
				case 8:	/* grey scale, read and triplicate. */
					r = g = b = l = getc (fin);
					a = 255;
					break;
					
				case 15:	/* 5 bits each of red green and blue. */
					/* watch byte order. */
					j = getc (fin);
					k = getc (fin);
					l = ((unsigned int)k << 8) + j;
					r = (byte)(((k & 0x7C) >> 2) << 3);
					g = (byte)((((k & 0x03) << 3) + ((j & 0xE0) >> 5)) << 3);
					b = (byte)((j & 0x1F) << 3);
					a = 255;
					break;
					
				case 16:	/* 5 bits each of red green and blue, 1 alpha bit. */
					/* watch byte order. */
					j = getc (fin);
					k = getc (fin);
					l = ((unsigned int)k << 8) + j;
					r = (byte)(((k & 0x7C) >> 2) << 3);
					g = (byte)((((k & 0x03) << 3) + ((j & 0xE0) >> 5)) << 3);
					b = (byte)((j & 0x1F) << 3);
					a = (k & 0x80) ? 255 : 0;
					break;
					
				case 24:	/* 8 bits each of blue, green and red. */
					b = getc (fin);
					g = getc (fin);
					r = getc (fin);
					a = 255;
					l = 0;
					break;
					
				case 32:	/* 8 bits each of blue, green, red and alpha. */
					b = getc (fin);
					g = getc (fin);
					r = getc (fin);
					a = getc (fin);
					l = 0;
					break;
					
				default:
					Console_Printf ("%s: Malformed TGA image: Illegal pixel_size '%d'\n", errhelp, pixel_size);
					data = Memory_free (data);
					if (mapped)
						ColorMap = Memory_free (ColorMap);
					return NULL;
			}
			
		PixEncode:
			if (mapped)
			{
				map_idx = l * 4;
				*dst++ = ColorMap[map_idx+0];
				*dst++ = ColorMap[map_idx+1];
				*dst++ = ColorMap[map_idx+2];
				*dst++ = ColorMap[map_idx+3];
			}
			else
			{
				*dst++ = r;
				*dst++ = g;
				*dst++ = b;
				*dst++ = a;
			}
		}
		
		if (interleave == TGA_IL_Four)
			truerow += 4;
		else if (interleave == TGA_IL_Two)
			truerow += 2;
		else
			truerow++;
		if (truerow >= h)
			truerow = ++baserow;
	}
	
	if (mapped)
		ColorMap = Memory_free (ColorMap);
	
	*didSucceed = True;
	return data;
}
#if 0
ftp://ftp.netbsd.org/pub/NetBSD/NetBSD-current/src/sys/dev/stbi/stb_image.c

static stbi_uc *tga_load(stbi *s, int *x, int *y, int *comp, int req_comp)
{
	//   read in the TGA header stuff
	int tga_offset = get8u(s);
	int tga_indexed = get8u(s);
	int tga_image_type = get8u(s);
	int tga_is_RLE = 0;
	int tga_palette_start = get16le(s);
	int tga_palette_len = get16le(s);
	int tga_palette_bits = get8u(s);
	int tga_x_origin = get16le(s);
	int tga_y_origin = get16le(s);
	int tga_width = get16le(s);
	int tga_height = get16le(s);
	int tga_bits_per_pixel = get8u(s);
	int tga_inverted = get8u(s);
	//   image data
	unsigned char *tga_data;
	unsigned char *tga_palette = NULL;
	int i, j;
	unsigned char raw_data[4];
	unsigned char trans_data[4];
	int RLE_count = 0;
	int RLE_repeating = 0;
	int read_next_pixel = 1;
	
	//   do a tiny bit of precessing
	if ( tga_image_type >= 8 )
	{
		tga_image_type -= 8;
		tga_is_RLE = 1;
	}
	/* int tga_alpha_bits = tga_inverted & 15; */
	tga_inverted = 1 - ((tga_inverted >> 5) & 1);
	
	//   error check
	if ( //(tga_indexed) ||
		(tga_width < 1) || (tga_height < 1) ||
		(tga_image_type < 1) || (tga_image_type > 3) ||
		((tga_bits_per_pixel != 8) && (tga_bits_per_pixel != 16) &&
		 (tga_bits_per_pixel != 24) && (tga_bits_per_pixel != 32))
		)
	{
		return NULL;
	}
	
	//   If I'm paletted, then I'll use the number of bits from the palette
	if ( tga_indexed )
	{
		tga_bits_per_pixel = tga_palette_bits;
	}
	
	//   tga info
	*x = tga_width;
	*y = tga_height;
	if ( (req_comp < 1) || (req_comp > 4) )
	{
		//   just use whatever the file was
		req_comp = tga_bits_per_pixel / 8;
		*comp = req_comp;
	} else
	{
		//   force a new number of components
		*comp = tga_bits_per_pixel/8;
	}
	tga_data = (unsigned char*)MALLOC( tga_width * tga_height * req_comp );
	
	//   skip to the data's starting position (offset usually = 0)
	skip(s, tga_offset );
	//   do I need to load a palette?
	if ( tga_indexed )
	{
		//   any data to skip? (offset usually = 0)
		skip(s, tga_palette_start );
		//   load the palette
		tga_palette = (unsigned char*)MALLOC( tga_palette_len * tga_palette_bits / 8 );
		if (!getn(s, tga_palette, tga_palette_len * tga_palette_bits / 8 ))
			return NULL;
	}
	//   load the data
	trans_data[0] = trans_data[1] = trans_data[2] = trans_data[3] = 0;
	for (i=0; i < tga_width * tga_height; ++i)
	{
		//   if I'm in RLE mode, do I need to get a RLE chunk?
		if ( tga_is_RLE )
		{
			if ( RLE_count == 0 )
			{
				//   yep, get the next byte as a RLE command
				int RLE_cmd = get8u(s);
				RLE_count = 1 + (RLE_cmd & 127);
				RLE_repeating = RLE_cmd >> 7;
				read_next_pixel = 1;
			} else if ( !RLE_repeating )
			{
				read_next_pixel = 1;
			}
		} else
		{
			read_next_pixel = 1;
		}
		//   OK, if I need to read a pixel, do it now
		if ( read_next_pixel )
		{
			//   load however much data we did have
			if ( tga_indexed )
			{
				//   read in 1 byte, then perform the lookup
				int pal_idx = get8u(s);
				if ( pal_idx >= tga_palette_len )
				{
					//   invalid index
					pal_idx = 0;
				}
				pal_idx *= tga_bits_per_pixel / 8;
				for (j = 0; j*8 < tga_bits_per_pixel; ++j)
				{
					raw_data[j] = tga_palette[pal_idx+j];
				}
			} else
			{
				//   read in the data raw
				for (j = 0; j*8 < tga_bits_per_pixel; ++j)
				{
					raw_data[j] = get8u(s);
				}
			}
			//   convert raw to the intermediate format
			switch (tga_bits_per_pixel)
			{
				case 8:
					//   Luminous => RGBA
					trans_data[0] = raw_data[0];
					trans_data[1] = raw_data[0];
					trans_data[2] = raw_data[0];
					trans_data[3] = 255;
					break;
				case 16:
					//   Luminous,Alpha => RGBA
					trans_data[0] = raw_data[0];
					trans_data[1] = raw_data[0];
					trans_data[2] = raw_data[0];
					trans_data[3] = raw_data[1];
					break;
				case 24:
					//   BGR => RGBA
					trans_data[0] = raw_data[2];
					trans_data[1] = raw_data[1];
					trans_data[2] = raw_data[0];
					trans_data[3] = 255;
					break;
				case 32:
					//   BGRA => RGBA
					trans_data[0] = raw_data[2];
					trans_data[1] = raw_data[1];
					trans_data[2] = raw_data[0];
					trans_data[3] = raw_data[3];
					break;
			}
			//   clear the reading flag for the next pixel
			read_next_pixel = 0;
		} // end of reading a pixel
		//   convert to final format
		switch (req_comp)
		{
			case 1:
				//   RGBA => Luminance
				tga_data[i*req_comp+0] = compute_y(trans_data[0],trans_data[1],trans_data[2]);
				break;
			case 2:
				//   RGBA => Luminance,Alpha
				tga_data[i*req_comp+0] = compute_y(trans_data[0],trans_data[1],trans_data[2]);
				tga_data[i*req_comp+1] = trans_data[3];
				break;
			case 3:
				//   RGBA => RGB
				tga_data[i*req_comp+0] = trans_data[0];
				tga_data[i*req_comp+1] = trans_data[1];
				tga_data[i*req_comp+2] = trans_data[2];
				break;
			case 4:
				//   RGBA => RGBA
				tga_data[i*req_comp+0] = trans_data[0];
				tga_data[i*req_comp+1] = trans_data[1];
				tga_data[i*req_comp+2] = trans_data[2];
				tga_data[i*req_comp+3] = trans_data[3];
				break;
		}
		//   in case we're in RLE mode, keep counting down
		--RLE_count;
	}
	//   do I need to invert the image?
	if ( tga_inverted )
	{
		for (j = 0; j*2 < tga_height; ++j)
		{
			int index1 = j * tga_width * req_comp;
			int index2 = (tga_height - 1 - j) * tga_width * req_comp;
			for (i = tga_width * req_comp; i > 0; --i)
			{
				unsigned char temp = tga_data[index1];
				tga_data[index1] = tga_data[index2];
				tga_data[index2] = temp;
				++index1;
				++index2;
			}
		}
	}
	//   clear my palette, if I had one
	if ( tga_palette != NULL )
	{
		FREE( tga_palette );
	}
	//   the things I do to get rid of an error message, and yet keep
	//   Microsoft's C compilers happy... [8^(
	tga_palette_start = tga_palette_len = tga_palette_bits =
	tga_x_origin = tga_y_origin = 0;
	//   OK, done
	return tga_data;
}

// http://nothings.org/stb_image.c

#endif
