//**********************************************************************************************************\
//
//	main.c:	Application entry point
//
//**********************************************************************************************************/
#include "project.h"

#define RESAMPLE_PERCENT 0.50f



/////////////////////////
typedef struct
{
	int	images; // For all images, one of these (and ONLY one) happens ...
	 int iNormals;
	 int iSpecials;
	 int iToolongs;
	 int iErrors;

	int TGA; // Packed + Unchanged + Ignored = num TGA.  
	 int tIgnored;
	 int tWadded;
	 int tPacked;
	 int tUnchanged;

	 int tRemovals;

//	int actions	;
	 int aRenames;
	 int aResizes;
	 int aFalseAlpha;
	 int aIgnoredAlpha;
	 
	int pChangeless;
} myStats_t;

FILE* actionFileList;

myStats_t	globalStats;
myStats_t	localStats;

#define STATPLUS(stat) { localStats.stat ++; globalStats.stat ++; } 

void ProcessFile (const char* currentFile,			// c:\quake\id1\textures\thiswad\#water.tga
				  const char* textureName,			// textures/#water.tga
				  const char* inPackName,			// *water
				  const char* tempFileForFolder,	
				  void* myWad2, 
				  void* myWad3, 
				  void* myPack,
				  void* myMasterPak)
{
// Muck with the image, do warnings, put in pack
	char*		bareFileName = String_SkipPath (currentFile);
	char		timeStamp[7] = "TM";
	int			peekWidth = 0, peekHeight = 0, peekBitsPerPixel = 0;
	int			currentFileSeconds = File_Time (currentFile);
	fbool		specialTexture = WadFile_Misc_Is_Name_Special (bareFileName);	// Pak it only.  Maybe resize
// 

	imageRGBA_t	myImage;

	STATPLUS (TGA);  // A file.
	STATPLUS (images);

	if (specialTexture) 
		STATPLUS (iSpecials) // No semi needed

	Image_Query_File_Dimensions (currentFile, &peekWidth, &peekHeight, &peekBitsPerPixel);
	Image_Import_FileX			(&myImage, currentFile);

// 2A. Special textures like _glow DO get a potential resize.  But they do not get alpha checked.  They do not go in a WAD.			
// 3A.  Detect unwanted presence of alpha channel.  Do not do for special textures.  Do not do for "{" prefix textures.
//		If we have an unwanted alpha channel, check and see if it is false.

	if (specialTexture == False && peekBitsPerPixel > 24 && bareFileName[0] != '{')
	{
		// Possibly improperly named texture
		if (Image_Info_AlphaType(&myImage) == COLORINFO_ALPHA_SOLID_WHITE)
		{
			printf ("Detected unexpected alpha channel (solid, so superfluous false alpha) ...");
			STATPLUS (aFalseAlpha) // No semi needed 
		}
		else
		{
			printf ("Detected valid alpha channel in a non-alpha texturename (should prefix with \"{\") ...");
			STATPLUS (aIgnoredAlpha) // No semi needed 
		}
	}
				
// 3B.	Resize if necessary to 32 multiple.  We do resize _bump and such textures too, to keep same size as regular texture counterpart (just because).
//		If we resize the TGA, we must save and get new timestamp.  Timestamp is vital for any packed file for FastEdit speed with ignoring TGAs.
//		Our 50% rescale depends on the source being divisible by 32.

	if (IS_MULTIPLE_OF_32(myImage.width) == False ||  IS_MULTIPLE_OF_32(myImage.height) == False)
	{
		// Trouble.  Enlarge canvas.  Write TGA back to disk with bright green background.  Note: INTEGER DIVISION
		int newWidth  = IS_MULTIPLE_OF_32(myImage.width ) == False ? Next_Highest_Multiple_Of (myImage.width, 32)  : myImage.width;
		int newHeight = IS_MULTIPLE_OF_32(myImage.height) == False ? Next_Highest_Multiple_Of (myImage.height, 32) : myImage.height;
		printf ("%ix%i size not mult of 32, padding to %i x %i ... ", myImage.width, myImage.height, newWidth, newHeight);
		Image_Enlarge_Canvas (&myImage, newWidth, newHeight, COLOR4BYTE_PUKEYELLOW);
		Image_Export_TGA_File_With_Header (&myImage, currentFile, "Padded Yellow", strlen("Padded Yellow")+1 );

		// MUST REFRESH THE TIME STAMP
		currentFileSeconds = File_Time (currentFile);

		STATPLUS(aResizes)
	}

// Build timestamp byte buffer.  Save the TGA with it in the header.  Put that TGA in the pack.
	memcpy								(&timeStamp[2], &currentFileSeconds, sizeof(currentFileSeconds));
	Image_Export_TGA_File_With_Header	(&myImage, tempFileForFolder, timeStamp, sizeof(timeStamp) );
	PakFile_Pack_File_As				(myPack, tempFileForFolder, inPackName);

// 3C.	Make MIP file.  Put in WAD
	if (specialTexture == False)
	{
		Image_Resample_By_Percent		(&myImage, RESAMPLE_PERCENT);

		Image_Export_Mip_File			(&myImage, tempFileForFolder, textureName);
		WadFile_Pack_File_As			(myWad2,   tempFileForFolder, textureName);
		WadFile_Pack_File_As			(myWad3,   tempFileForFolder, textureName);

		STATPLUS (iNormals)	// specials, toolongs, errors <--- they don't get wadded
		STATPLUS (tWadded)
	}

	Image_Destroy						(&myImage);
	
	
	
	STATPLUS (tPacked)

	if (myMasterPak)
		PakFile_Pack_File_As			(myMasterPak, tempFileForFolder /* myTGAfiles[j].text*/, inPackName);

			
}

void ProcessFolder (const char* folderName, const char* quakeGameDirFolder, void *myMasterPak, fbool isRebuild)
{
// First what is the folder name
	char*		BareFolderName = String_SkipPath (folderName);
// Get a list of TGAs in the folder
	int			numMyTGAFiles;
	text1024_t* myTGAfiles	= FileMatch_Populate_Alloc   (folderName, "*.tga", &numMyTGAFiles);

	if (numMyTGAFiles == 0) return; // Just leave

	memset (&localStats, 0, sizeof(localStats));

	do
	{
// Build list of all output paks and wads
		char _Wad2Name[MAX_PATH_IS_1024];   char* Wad2Name = String_Path (_Wad2Name, "%s/%s.wad",          quakeGameDirFolder, BareFolderName); // c:\textures\metals.wad
		char _Wad3Name[MAX_PATH_IS_1024];	
		char* Wad3Name = String_Path (_Wad3Name, "%s/%s.hlwad",        quakeGameDirFolder, BareFolderName); // c:\textures\metals.hlwad
		char _PakName[MAX_PATH_IS_1024];    char* PakName  = String_Path (_PakName,  "%s/textures_%s.pak", quakeGameDirFolder, BareFolderName); // c:\textures\textures_metals.pak
		char _TempFile[MAX_PATH_IS_1024];   char* TempFile = String_Path (_TempFile, "%s/%s",		       quakeGameDirFolder, "_temp.tmp"); 
// Pak Exists?
		fbool packExisted	= isRebuild ? False : File_Exists (PakName);
// In case we need these (we will unless the folder is empty
		void* myWad2		= isRebuild ? WadFile_Open  (Wad2Name, ARCHIVE_CREATE)		: WadFile_Auto_Open2 (Wad2Name);
		void* myWad3		= isRebuild ? WadFile_Open  (Wad3Name, ARCHIVE_CREATE_WAD3) : WadFile_Auto_Open3 (Wad3Name);
		void* myPack		= isRebuild ? PakFile_Open  (PakName,  ARCHIVE_CREATE)		: PakFile_Auto_Open  (PakName);
// We are going to mark every file we run across that isn't "too long".  Unmarked files need to be deleted.
		int			numMyPackedFiles = 0;
		text1024_t* myPackedFiles = packExisted ? PakFile_Directory_Alloc (myPack, &numMyPackedFiles) : NULL;
		fbool	existingPakDidNotChange = False;
		int		j;

		printf ("\nWAD2 Mode: %s\n", WadFile_Is_Edit_Mode (myWad2) ? "EDIT" : "CREATE");
		printf ("WAD3 Mode: %s\n", WadFile_Is_Edit_Mode (myWad3) ? "EDIT" : "CREATE");
		printf ("Pack Mode: %s\n\n", PakFile_Is_Edit_Mode (myPack) ? "EDIT" : "CREATE");


		for (j = 0; j < numMyTGAFiles; j ++)
		{
			char*	bareFileName = String_SkipPath (myTGAfiles[j].text);
			char	textureName[256];
			char	inPackName[56];
			int		timeStampTGA = File_Time (myTGAfiles[j].text);
			int		k;
			
			printf ("\rFolder  %-10s -> File (%i/%i): %-19s ", BareFolderName, j + 1, numMyTGAFiles, bareFileName);
			if (j + 1 != numMyTGAFiles)
				fflush(stdout);
			else
				printf ("\n");

			if (StringStartsWithCaseLess(bareFileName, "_toolong15_"))
			{
				STATPLUS  (iToolongs)
				STATPLUS (tIgnored)
				continue; // Don't even bother
			}

			// Generate textureName
			StringLCopy (textureName, bareFileName);
			String_Edit_StripExtension (textureName);
			SNPrintf (inPackName, sizeof(inPackName), "textures/%s", bareFileName); 

			if (textureName[0] == '#')	// Liquid texture name parallel
				textureName[0] = '*';

			if (WadFile_Misc_Is_Texture_Name_Too_Long (textureName) )
			{
				STATPLUS (iToolongs)
				STATPLUS (tIgnored)
				STATPLUS (aRenames)
				File_Rename_With_Prefix (myTGAfiles[j].text, "_toolong15_");
				continue; // Don't even bother
			}

// Mark the file because if we reach this point we ARE working with valid files now
			for (k = 0; k < numMyPackedFiles; k ++)
			{
				char* thisName = String_SkipPath(myPackedFiles[k].text);
				if (StringMatch(thisName, bareFileName) )
				{
					myPackedFiles[k].marked = True;
					break;
				}
				//printf ("File %s isn't in pack?\n", thisName);
			}

// Back the business like usual
			if (PakFile_Is_Edit_Mode (myPack)) // Pointless if mode is CREATE
			{
				if (PakFile_File_TGA_Is_Packed (myPack, inPackName, timeStampTGA))
				{
					//printf ("timestamp match; skipping.");
					STATPLUS (tUnchanged)
					continue; // Nothing happened.
				}
				else if (PakFile_Is_Edit_Mode (myPack))
					printf ("Says this file has changed?\n %s", inPackName);
			}
			ProcessFile (myTGAfiles[j].text, textureName, inPackName, TempFile, myWad2, myWad3, myPack, myMasterPak);

		}

		// Delete.  Possibly.  Look through unmarked pak entries.
		if (packExisted && isRebuild == False) // Which assumes we are in edit mode
		{
			for (j = 0; j < numMyPackedFiles; j ++)
			{
				if (myPackedFiles[j].marked == False && myPackedFiles[j].text[0])
				{
					// File exists in pak.  But not in directory.  Delete it from archives.
					char* currentInPackName = myPackedFiles[j].text;
					char* textureName		= String_SkipPath (currentInPackName);
					printf ("File in pack %s seems to be deleted in folder\n", currentInPackName);
					PakFile_Delete_File (myPack,  currentInPackName);
					
					STATPLUS (tRemovals);

					if (myMasterPak)
						WadFile_Delete_File (myMasterPak, currentInPackName); // This could be problematic. Nothing says folders can't have same texture names

					if (textureName[0] == '#')	// Liquid texture name parallel
						textureName[0] = '*';
// Kill extension.
					String_Edit_StripExtension (textureName); // Let's ruin the data even more 

					WadFile_Delete_File (myWad2, textureName);
					WadFile_Delete_File (myWad3, textureName);
				}
			}
		}
		// CLOSE

		existingPakDidNotChange = PakFile_Is_Edit_Mode (myPack) && PakFile_NumChanges (myPack) == 0;

		PakFile_Close (myPack);
		WadFile_Close (myWad3);
		WadFile_Close (myWad2);


		// MAYBE COMPRESS:  If we were in EDIT mode and no changes occurred, assume we want to compress.

		if (existingPakDidNotChange)
		{
			STATPLUS (pChangeless);
			//printf ("Compressing ...\n");
			//WadFile_Compress (Wad2Name);
			//WadFile_Compress (Wad3Name);
			//WadFile_Compress (PakName);
			//printf ("Compression complete.\n");
			printf ("Not compressing.  But we should.\n");
		}

		if (myPackedFiles) Memory_free (myPackedFiles);
	} while (0);

	Memory_free (myTGAfiles);
}





///////////////////////////////////////////////////////////////////////////////////
// BEGIN
///////////////////////////////////////////////////////////////////////////////////

void main_go (void)
{
	char		tempHelp[MAX_PATH_IS_1024];
	fbool		canWriteIn;
	fbool		canWriteOut;
	fbool		isRebuild = False;
//
// INPUT
//
#ifdef _DEBUG
	char*		inputDirectory = currentWorkingDirectory;
#else
	char*		inputDirectory = executableDirectory;
#endif	
	int			numMyFolders, i, j;
	text1024_t*	myFolders = FolderMatch_Populate_Alloc (inputDirectory, "*", &numMyFolders);

//
// MASTER PAK
//
	char*		masterPakName = "pak2.pak";
	char		masterPakPath[MAX_PATH_IS_1024] = "\0";
	void*		myMasterPak = NULL;
//
// OUTPUT
//
	char		quakeGameDirFolder[MAX_PATH_IS_1024] = "\0"; // OUTPUT FOLDER
	FILE*		fout; // LOG ONLY TELLS US ABOUT TEXTURES.  NOT GENERAL BOREDOM STUFF.  Someone can pipe to an output file if they want.
	fbool		verbose = CheckParm("-verbose");
#define verbose_print(msg) if (verbose) printf (msg)
// 
// OPTIONS
//

// 1. "Help option"

	if (CheckParm ("-help") || CheckParm ("/h"))
	{
		printf ("TextureAI - May 24, 2012 by Baker @ QuakeOne.com\n");
		printf ("\n");
		printf ("Allows mapping using traditional Quake map tools\n");
		printf ("in a way intended for using replacement textures.\n");
		printf ("\n");
		printf ("Options:\n");
		printf ("\n");
		printf ("-rebuild			 Rebuilds fresh (slow)\n");
		printf ("\n");
		printf ("-noseek			 Writes output where this binary is\n");
		printf ("-outputdir <path>   Manually set output directory\n");
		printf ("-game <gamedir>     Output to gamedir folder not id1\n");
		printf ("\n");		
		printf ("-nopak              Don't create a master pak\n");
		printf ("-masterpak <file>   Set master pak name (pak2.pak)\n");
		printf ("\n");
		printf ("-verbose            Print more detailed information.\n");
		printf ("-help               Display this help information.\n");
		printf ("\n");
		printf ("Note: a master pak is a the aggregate collection of\n");
		printf ("all high resolution TGA textures in a single archive.\n");
		exit (0);
	}

// 2. Determine output directory.  This can be a bit involved.

	if (CheckParm ("-rebuild"))
	{
		isRebuild = True;
		printf ("-rebuild detected\n");
	}

	if (CheckParm ("-noseek"))
	{
		StringLCopy (quakeGameDirFolder, inputDirectory);
		printf ("-noseek detected\n");
		goto outdir_done;
	}

	if (CheckParm ("-outputdir") && CheckParm ("-outputdir") + 1 < commandLineArgumentsCount)
	{
		// Let's hope they don't mess it up.
		StringLCopy (quakeGameDirFolder, commandLineArguments[CheckParm ("-outputdir") + 1] );
		printf ("-outputdir detected: %s\n", quakeGameDirFolder);
		goto outdir_done;
	}

// 2a. Find where we are outputing.  Yuck!  But it IS worth it absolutely!

	do 
	{
		char		lookUpPathFromHereTemp[MAX_PATH_IS_1024];
// 2b. Determine the gamedir we are looking for.  Normally id1.  Can be overriden with -game parameter
		char		gameDirWanted[MAX_PATH_IS_1024] = "id1";

		if (CheckParm("-game") && CheckParm("-game") + 1 < commandLineArgumentsCount)
		{
			StringLCopy (gameDirWanted, commandLineArguments[CheckParm("-game") + 1] );
			printf ("-game parameter detected: '%s'\n", gameDirWanted);
		}
// 2c. Perform the seek for id1 folder up the dir path
		SNPrintf (lookUpPathFromHereTemp, sizeof(lookUpPathFromHereTemp), "%s/", inputDirectory);  

		do
		{
			String_Edit_To_Parent_Path (lookUpPathFromHereTemp);  // Strip another level off directory
			SNPrintf	(tempHelp, sizeof(tempHelp), "%s/%s/pak0.pak", lookUpPathFromHereTemp, gameDirWanted);

			// See if we found the folder we were looking for, if not have we hit the "end of the road"
				 if (File_Exists (tempHelp))						SNPrintf (quakeGameDirFolder, sizeof(quakeGameDirFolder), "%s/%s", lookUpPathFromHereTemp, gameDirWanted);
			else if (String_isPathTopMost (lookUpPathFromHereTemp)) StringLCopy (quakeGameDirFolder, inputDirectory);  // No dice, use default input directory

		} while (quakeGameDirFolder[0] == 0);
		// End of name hunt
	} while (0);

outdir_done:
	printf ("\nOutput directory: %s\n", quakeGameDirFolder);

// 3.  Check ability to write to input and output dirs
	
	if ( (canWriteIn  = File_Directory_TestWrite (inputDirectory))     == False)	printf ("Error: Cannot write to input directory: %s\n", inputDirectory);
	if ( (canWriteOut = File_Directory_TestWrite (quakeGameDirFolder)) == False)	printf ("Error: Cannot write to output directory: %s\n", quakeGameDirFolder);

	if (canWriteIn == False || canWriteOut == False) exit (1);

// 4. Log file.


// 5. Determine master pak using options
	
	if (CheckParm ("-nopak"))
	{
		printf ("-nopak option detected.\n\n");
		masterPakName = NULL; // Do not make
	}
	else if (CheckParm ("-masterpakname") && CheckParm ("-masterpakname") + 1 < commandLineArgumentsCount)
	{
		masterPakName = commandLineArguments [CheckParm ("-masterpakname") + 1];
		printf ("-masterpakname option detected: '%s'\n\n", masterPakName);
	}

	if (masterPakName)
		SNPrintf (masterPakPath, sizeof(masterPakPath), "%s/%s", quakeGameDirFolder, masterPakName);

// 6. If using a master pak, open for writing and determine if creating or fastediting.
	if (masterPakName)
		if (isRebuild)
			myMasterPak = PakFile_Open (masterPakPath, ARCHIVE_CREATE);
		else
			myMasterPak = PakFile_Auto_Open (masterPakPath); // CREATE OR FASTEDIT depending on situation

	printf ("Number of folders to process = %i\n", numMyFolders);

	// Go through each folder.  Make WAD2, WAD3, PAK (plus add to super pak)
	for (i = 0; i < numMyFolders; i ++)
		ProcessFolder (myFolders[i].text, quakeGameDirFolder, myMasterPak, isRebuild);
	
	if (myMasterPak)
	{	// Do a rebuild if isrebuild is false (no need, isrebuild does create) edit mode is true (create doesn't need rebuild) and no changes made
		fbool doCompress = isRebuild == False && PakFile_Is_Edit_Mode(myMasterPak) == True && PakFile_NumChanges (myMasterPak) == 0;
		PakFile_Close (myMasterPak);
		// If nothing changed at rebuild the entire pak
		if (doCompress)
			PakFile_Compress (masterPakName);
	}
	
	if (myFolders) Memory_free (myFolders);

	///////////////////////////////////////////////////////////////////////////////////
	// END
	///////////////////////////////////////////////////////////////////////////////////

}

