//**********************************************************************************************************\
//
//	strings.c:	String functions, etc.
//
//**********************************************************************************************************/

#include "project.h"

void String_Edit_SlashesForward_Like_Unix (char *WindowsStylePath)
{
	size_t	i;
	// Translate "\" to "/"
	for (i = 0 ; i < strlen(WindowsStylePath) ; i ++)
		if (WindowsStylePath[i] == '\\')
			WindowsStylePath[i] = '/';
}

void String_Edit_To_Parent_Path (char* myPath)
{
	char* terminatePoint = strrchr (myPath, '/');

	if (terminatePoint)
		*terminatePoint = '\0';

}

void String_Edit_StripExtension (char* myEditString)
{
	size_t cursor = strlen(myEditString) - 1;
	for ( ; cursor > 0; cursor --)
        if (myEditString[cursor] == '.')
        {
            myEditString[cursor] = 0; // NULL terminate there
            break;
        }
}

void String_Edit_ChangeExtension (char* myEditString, const char* newExtension)
{
	String_Edit_StripExtension (myEditString);
	String_Edit_Append_To_Name (myEditString, newExtension);
}

char* String_Edit_Append_To_Name (char* filename, const char* appendToName) 
{
	char	tempbuffer[MAX_PATH_IS_1024];			// We need this for string work
	char*	extension = strrchr (filename, '.');	// Locate extension (NULL if none)

	StringLCopy	(tempbuffer, filename);
	
	if (extension)	String_Edit_StripExtension	(tempbuffer);
					StringLCat					(tempbuffer,appendToName);
	if (extension)	StringLCat					(tempbuffer, extension); // Tack the extension back on

	// Move the modified string back into the filename, this is a String_Edit function you know ...
	strcpy	(filename, tempbuffer); // Can't StringLCopy, length of filename is "unknown" here

	return filename;
}




int File_HasDotExtension (const char *filename, const char *extension)
{
	// Check to ensure strlen is equal or exceeds strlen extension
	if (strlen(filename) < strlen(extension))	// This allows for stupid "situations like filename is '.dem'
		return 0;

	// Now check for match
	//                       Go to end of the file and move back length of extension vs. extension
	if (StringMatchCaseless (filename + strlen(filename) - strlen(extension), extension))
		return 1; // Match!

	// No Match
	return 0;
}


char* String_SkipPath (const char *pathName)
{
	char*	lastSlash = strrchr(pathName, '/');

	if (lastSlash == NULL) return (char*)pathName; // No slash found, return whole file name

	return lastSlash + 1; // Return first character after the slash
}


fbool StringEndsWith (const char* bigString, const char* suffix)
{
	int len1 = strlen(bigString);
	int len2 = strlen(suffix);

	if ((len1 - len2) < 0)								return False; // String doesn't even fit
	if (strcmp(&bigString[len1 - len2], suffix) == 0)	return True;  // Match

	return False;
}

fbool String_isPathTopMost (const char *pathName)
{
	// a string like textures/something/whatever
	// a string like c:/quake/id1

	// If string has a colon and then a slash, start after that.
	
	const char* firstColon	= strchr (pathName, ':');
	const char* pathStart	= firstColon ? firstColon + 2 : pathName;
	const char* foundSlash	= strchr (pathStart, '/');

	if (foundSlash)
		return False; // Has Parent so IS NOT TOPMOST
	else
		return True;
}

char*	String_Buffer_Format					(char* buffer, size_t bufferSize, const char* formatText, const char* path, const char* filename)
{
	SNPrintf (buffer, bufferSize, formatText, path, filename);
	return buffer;
}